package de.unirostock.sems.XmlFileServerClient;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;

import de.unirostock.sems.XmlFileServerClient.exceptions.ModelAlreadyExistsException;
import de.unirostock.sems.XmlFileServerClient.exceptions.ModelNotFoundException;
import de.unirostock.sems.XmlFileServerClient.exceptions.UnsupportedUriException;
import de.unirostock.sems.XmlFileServerClient.exceptions.XmlFileServerBadRequestException;
import de.unirostock.sems.XmlFileServerClient.exceptions.XmlFileServerProtocollException;

/**
 * Interface for communication with a XmlFileServer. Independent from the communication channel
 * 
 * @author Martin Peters
 *
 */
public interface XmlFileServer {
	public static final String URL_ENCODING = "UTF-8";

	/**
	 * Resolves the model URI and returns the modelSource
	 * 
	 * @param model URI
	 * @return InputStream or {@code NULL} if URI not found
	 * @throws FileNotFoundException 
	 * @throws UnsupportedUriException 
	 * @throws XmlFileServerProtocollException 
	 * @throws IOException 
	 */
	public InputStream resolveModelUri( URI model ) throws ModelNotFoundException, UnsupportedUriException, XmlFileServerProtocollException, IOException;

	/**
	 * Checks if the model referenced by the URI exits and is available
	 * 
	 * @param model
	 * @return boolean
	 * @throws UnsupportedUriException 
	 * @throws IOException 
	 * @throws XmlFileServerProtocollException 
	 */
	public boolean exist( URI model ) throws UnsupportedUriException, IOException, XmlFileServerProtocollException;

	/**
	 * Checks if the URI is resolvable by the FileServer. <br>
	 * (Tests only scheme and host)
	 * 
	 * @param model
	 * @return
	 */
	public boolean isResolvableUri(URI model);

	/**
	 * Pushes a new model (-Version) into the storage. <br>
	 * Returns the URI to the new model or <code>null</code> if failed.
	 * 
	 * @param modelId
	 * @param versionId
	 * @param modelSource
	 * @return URI to the new model
	 * @throws IOException 
	 * @throws UnsupportedUriException 
	 * @throws XmlFileServerProtocollException 
	 * @throws XmlFileServerBadRequestException 
	 * @throws ModelAlreadyExistsException 
	 */
	public URI pushModel( String modelId, String versionId, InputStream modelSource ) throws IOException, UnsupportedUriException, XmlFileServerProtocollException, ModelAlreadyExistsException, XmlFileServerBadRequestException;
	
	/**
	 * Pushes a new model (-Version) into the storage. <br>
	 * Returns the URI to the new model or <code>null</code> if failed.
	 * The modelId will be generated from the server, as well as the model URI
	 * 
	 * @param repositoryUrl
	 * @param filePath
	 * @param versionId
	 * @param modelSource
	 * @return URI to the new model
	 * @throws UnsupportedUriException 
	 * @throws IOException 
	 * @throws XmlFileServerProtocollException 
	 * @throws XmlFileServerBadRequestException 
	 * @throws ModelAlreadyExistsException 
	 */
	public URI pushModel( String repositoryUrl, String filePath, String versionId, InputStream modelSource ) throws IOException, UnsupportedUriException, XmlFileServerProtocollException, ModelAlreadyExistsException, XmlFileServerBadRequestException;
	
	/**
	 * Pushes a new model (-Version) into the storage. <br>
	 * Returns the URI to the new model or {@code null} if failed. <br><br>
	 * The link corresponds to following pattern:<br> {schema}://{host}/{repositoryUrl}/{versionId}/{fileUrl} <br>
	 * repositoryUrl and fileUrl together expresses the modelId
	 * 
	 * @param repositoryUrl
	 * @param fileUrl
	 * @param versionId
	 * @param modelSource
	 * @return URI to the new model
	 * @throws IOException
	 * @throws UnsupportedUriException
	 * @throws XmlFileServerProtocollException 
	 * @throws ModelAlreadyExistsException 
	 * @throws XmlFileServerBadRequestException 
	 */
	public URI pushModel(String modelId, String versionId, String repositoryUrl, String filePath, InputStream modelSource) throws IOException, UnsupportedUriException, XmlFileServerProtocollException, ModelAlreadyExistsException, XmlFileServerBadRequestException;
}
