package de.unirostock.sems.XmlFileServerClient.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.Charset;
import java.text.MessageFormat;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpHead;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.mime.MultipartEntity;
import org.apache.http.entity.mime.content.InputStreamBody;
import org.apache.http.entity.mime.content.StringBody;
import org.apache.http.impl.client.DefaultHttpClient;

import de.unirostock.sems.XmlFileServerClient.XmlFileServer;
import de.unirostock.sems.XmlFileServerClient.exceptions.ModelAlreadyExistsException;
import de.unirostock.sems.XmlFileServerClient.exceptions.ModelNotFoundException;
import de.unirostock.sems.XmlFileServerClient.exceptions.UnsupportedUriException;
import de.unirostock.sems.XmlFileServerClient.exceptions.XmlFileServerBadRequestException;
import de.unirostock.sems.XmlFileServerClient.exceptions.XmlFileServerProtocollException;

public class HttpXmlFileServerClient implements XmlFileServer {

	// TODO Logging
	private final Log log = LogFactory.getLog( HttpXmlFileServerClient.class );

	public static final int STATUS_NOT_FOUND = 404;
	public static final int STATUS_MODEL_ALREADY_EXISTS = 403;
	public static final int STATUS_BAD_REQUEST = 400;
	public static final int STATUS_SERVER_ERROR = 500;

	public static final int STATUS_MODEL_WRITTEN = 201;
	public static final int STATUS_MODEL_DELIVERED = 200;

	// POST Parameter and when the put request has'nt specified a path
	private static final String PARAMETER_MODEL_ID		= "modelId";
	private static final String PARAMETER_VERSION_ID	= "versionId";
	private static final String PARAMETER_REPO_URL		= "repositoryUrl";
	private static final String PARAMETER_FILE_PATH		= "filePath";
	private static final String PARAMETER_MODEL_STREAM	= "file";

	private static final String MODEL_MIMETYPE			= "application/xml";
	
	// Http Header Field for the generated URI
	private static final String HEADER_LOCATION = "location";
	
	private URI xmlFileServer;
	private String prefixPath;
	private HttpClient client;

	public HttpXmlFileServerClient( URI xmlFileServer ) {
		this.xmlFileServer = xmlFileServer;
		this.prefixPath = xmlFileServer.getPath();


		// creates a new a Http Client
		client = new DefaultHttpClient();
	}

	private URI generateRequestUri( URI model ) throws UnsupportedUriException {
		String requestPath;
		URI requestUri = null;

		if( prefixPath.isEmpty() || prefixPath.equals("/") ) {
			// no prefixPath used
			requestPath = model.getPath();
		}
		else {
			requestPath = model.getPath();
			if( !requestPath.toLowerCase().startsWith(prefixPath.toLowerCase()) ) {
				if( !prefixPath.endsWith("/") && !requestPath.startsWith("/") )
					requestPath = prefixPath + "/" + requestPath;
				else if( prefixPath.endsWith("/") && requestPath.startsWith("/") )
					requestPath = prefixPath.substring(1) + requestPath;
			}
		}

		try {
			requestUri = new URI( xmlFileServer.getScheme(), null, xmlFileServer.getHost(), xmlFileServer.getPort(), requestPath, null, null);
		} catch (URISyntaxException e) {
			String message = MessageFormat.format("UriSyntax error while generating the requestUri for {0}", model);
			log.error(message, e);
			throw new UnsupportedUriException(message, e);
		}

		return requestUri;
	}
	
	private String generateDebugModelIdentifier( String modelId, String filePath, String repositoryUrl, String versionId ) {
		StringBuilder result = new StringBuilder();
		if( modelId != null && !modelId.isEmpty() ) {
			result.append(modelId);
		}
		else {
			result.append(repositoryUrl);
			result.append(':');
			result.append(filePath);
		}
		result.append(':');
		result.append(versionId);
		
		return result.toString();
	}
	
	@Override
	public InputStream resolveModelUri(URI model) throws ModelNotFoundException, UnsupportedUriException, XmlFileServerProtocollException, IOException {
		HttpGet request;
		HttpResponse response = null;

		URI requestUri = generateRequestUri(model);

		// creates the Request
		request = new HttpGet(requestUri);
		// performs it
		try {
			response = client.execute(request);
		} catch (ClientProtocolException e) {
			// Error with the Http Protokoll
			String message = MessageFormat.format("Http Client Protocoll Exception while resolving model {0}", model);
			log.error(message, e);
			throw new XmlFileServerProtocollException(message, e);
		} catch (IOException e) {
			// some IO stuff went wrong...
			String message = MessageFormat.format("IOException while resolving model {0}", model);
			log.error(message, e);
			throw new IOException(message, e);
		}

		if( response == null ) {
			String message = MessageFormat.format("No response. {0}", model);
			log.error(message);
			return null;
		}

		if( response.getStatusLine().getStatusCode() == STATUS_MODEL_DELIVERED ) {
			try {
				return response.getEntity().getContent();
			} catch (IllegalStateException e) {
				String message = MessageFormat.format("IllegalState while returning the resolved data of model {0}", model);
				log.error(message, e);
				throw new IllegalStateException(message, e);
			} catch (IOException e) {
				String message = MessageFormat.format("IOException while returning the resolved data of model {0}", model);
				log.error(message, e);
				throw new IOException(message, e);
			}
		} else {
			throw new ModelNotFoundException( response.getStatusLine().getReasonPhrase() );
		}
	}

	@Override
	public boolean exist(URI model) throws UnsupportedUriException, IOException, XmlFileServerProtocollException {
		HttpHead request;
		HttpResponse response = null;
		URI requestUri = generateRequestUri(model);

		// create Request
		request = new HttpHead(requestUri);

		try {
			response = client.execute(request);
		} catch (ClientProtocolException e) {
			// Error with the Http Protokoll
			String message = MessageFormat.format("Http Client Protocoll Exception while resolving model {0}", model);
			log.error(message, e);
			throw new XmlFileServerProtocollException(message, e);
		} catch (IOException e) {
			// some IO stuff went wrong...
			String message = MessageFormat.format("IOException while resolving model {0}", model);
			log.error(message, e);
			throw new IOException(message, e);
		}

		if( response == null ) {
			String message = MessageFormat.format("No response. {0}", model);
			log.error(message);
			return false;
		}

		if( response.getStatusLine().getStatusCode() == STATUS_MODEL_DELIVERED ) {
			return true;
		}

		return false;
	}

	@Override
	public boolean isResolvableUri(URI model) {

		if( !model.getScheme().equals( xmlFileServer.getScheme() ))
			return false;

		if( !model.getHost().equals( xmlFileServer.getHost() ))
			return false;

		return true;
	}

	@Override
	public URI pushModel(String modelId, String versionId, InputStream modelSource) throws IOException, UnsupportedUriException, XmlFileServerProtocollException, ModelAlreadyExistsException, XmlFileServerBadRequestException {
		return pushModel(modelId, versionId, null, null, modelSource);
	}

	@Override
	public URI pushModel(String repositoryUrl, String filePath, String versionId, InputStream modelSource) throws IOException, UnsupportedUriException, XmlFileServerProtocollException, ModelAlreadyExistsException, XmlFileServerBadRequestException {
		return pushModel(null, versionId, repositoryUrl, filePath, modelSource);
	}

	@Override
	public URI pushModel(String modelId, String versionId, String repositoryUrl, String filePath, InputStream modelSource)
			throws UnsupportedUriException, XmlFileServerProtocollException, IOException, ModelAlreadyExistsException, XmlFileServerBadRequestException {
		
		if( versionId == null || versionId.isEmpty() ) {
			// the versionId have to be set, in every case!
			throw new IllegalArgumentException("The versionId is not set!");
		}
		if( modelSource == null ) {
			// no stream - no model - nothing todo
			// so why call this method? For punishment you get an exception! Tah!
			throw new IllegalArgumentException("The modelStream is null! So why did you call this method?");
		}

		HttpPost request;
		HttpResponse response = null;

		// Get the charset
		Charset charset = Charset.forName(URL_ENCODING);

		// ++++  START baking multipart entity ++++
		// Creating the Http Entity in Multipart form in default mode and generated Boundary, but with an specific encoding :)
		MultipartEntity entity = new MultipartEntity(null, null, charset );

		try {
			// adding version information
			entity.addPart(PARAMETER_VERSION_ID, new StringBody(versionId, charset) );

			// adding location information (rhymes, bitch!)
			if( repositoryUrl != null && filePath != null && !repositoryUrl.isEmpty() && !filePath.isEmpty() ) {
				// adds filePath and repositoryUrl as parameter
				entity.addPart( PARAMETER_REPO_URL, new StringBody(repositoryUrl, charset) );
				entity.addPart( PARAMETER_FILE_PATH, new StringBody(filePath, charset) );
			}
			else if( modelId != null && !modelId.isEmpty() ) {
				// add the modelId as parameter
				entity.addPart( PARAMETER_MODEL_ID, new StringBody(modelId, charset) );
			}
			else {
				// insufficient amount of parameter... bam... Exception!
				throw new IllegalArgumentException("Insufficient amount of parameter. You are supposed to either set modelId or repositoryUrl and filePath!");
			}

			// adding the stream
			entity.addPart( PARAMETER_MODEL_STREAM, new InputStreamBody(modelSource, MODEL_MIMETYPE, versionId) );
		}
		catch (UnsupportedEncodingException e) {
			String message = MessageFormat.format("Charset {0} is not supported!",  URL_ENCODING);
			log.fatal(message, e);
		}

		// ++++ END baking multipart entity ++++
		
		try {
			// create post request directly to the base dir
			request = new HttpPost(xmlFileServer);
			// set the entity
			request.setEntity( entity );
			
			// do it!
			response = client.execute(request);
		} catch (ClientProtocolException e) {
			// Error with the Http Protokoll
			String message = MessageFormat.format("Http Client Protocoll Exception while pushing model {0} to server", generateDebugModelIdentifier(modelId, filePath, repositoryUrl, versionId));
			log.error(message, e);
			throw new XmlFileServerProtocollException(message, e);
		} catch (IOException e) {
			// some IO stuff went wrong...
			String message = MessageFormat.format("IOException while pushing model {0} to server", generateDebugModelIdentifier(modelId, filePath, repositoryUrl, versionId));
			log.error(message, e);
			throw new IOException(message, e);
		}
		
		request.releaseConnection();
		
		int statusCode = response.getStatusLine().getStatusCode();
		if( statusCode != STATUS_MODEL_WRITTEN ) {
			// writing the model was not successful
			
			String status = response.getStatusLine().getReasonPhrase();
			
			// checking all know ExceptionIDs
			if( status.endsWith(ModelAlreadyExistsException.ID) ) {
				// Model already existed
				String message = MessageFormat.format("Model {1} already exist: \"{0}\"", status, generateDebugModelIdentifier(modelId, filePath, repositoryUrl, versionId));
				log.error(message);
				throw new ModelAlreadyExistsException(message);
			}
			else if( status.endsWith(UnsupportedUriException.ID) ) {
				// the URI is not supported
				String message = MessageFormat.format("Unsupported URI while pushing model {1} : \"{0}\"", status, generateDebugModelIdentifier(modelId, filePath, repositoryUrl, versionId));
				log.error(message);
				throw new UnsupportedUriException(message);
			}
			else {
				// some other bad Request
				String message = MessageFormat.format("BadRequest while pushing model {1} : \"{0}\"", status, generateDebugModelIdentifier(modelId, filePath, repositoryUrl, versionId));
				log.error(message);
				throw new XmlFileServerBadRequestException(message);
			}
		}
		else {
			// everything went right
			// the new URI should be in the location-header-field
			
			Header location = response.getLastHeader(HEADER_LOCATION);
			// empty location field
			if( location == null ) {
				log.error( MessageFormat.format("Location field is null after pushing model {0}", generateDebugModelIdentifier(modelId, filePath, repositoryUrl, versionId)) );
				return null;
			}
			
			try {
				return new URI( location.getValue() );
			} catch (URISyntaxException e) {
				String message = MessageFormat.format("Can not parse the location field after pushed model {0} to server", generateDebugModelIdentifier(modelId, filePath, repositoryUrl, versionId)); 
				log.error( message );
				throw new UnsupportedUriException(message, e);
			}
			
		}
	}



}
